const {By, Button, until,WebElement} = require("selenium-webdriver");
const crypto = require('crypto');
const fs = require('fs');
const path = require('path');
const sharp = require('sharp');

exports.Commander = class Commander {
	constructor(driver) {
		this.driver = driver;
		this.element = null;
	}

	async get(url) {
		await this.driver.get(url);
	}

	async resetMousePosition() {
		await this.driver.get("webdriver:///mouse/position/reset")
	}

	parseQuery(query) {
		if (typeof query === "string") {
			let q = null;
			if (query.startsWith("#")) {
				q = By.id(query.substring(1));
			} else if (query.startsWith(".")) {
				q = By.className(query.substring(1));
			} else if (query.startsWith("+")) {
				q = By.name(query.substring(1));
			} else if (query.startsWith("<")) {
				q = By.xpath(query.substring(1));
			} else {
				throw new Error("bad query " + query + ".");
			}
			return q;
		}
		return query;
	}
	isPosition(d) {
		return d && d.x !== undefined && d.y !== undefined;
	}
	isToPosition(d) {
		return d && d.tx !== undefined && d.ty !== undefined;
	}
	isRect(d) {
		return d && d.left !== undefined && d.top !== undefined && d.width !== undefined && d.height !== undefined;
	}

	delay(msec) {
		return new Promise(resolve => setTimeout(resolve, msec));
	}

	async queryElement(query) {
		if(query === null || query === undefined ) {
			if (this.element !== null)
				return this.element;
		}
		if(query instanceof WebElement) {
			this.element = query;
			return query;
		}
		let elm = null;
		try {
			var q = this.parseQuery(query);
			elm = await this.driver.findElement(q);
		}
		catch(e) {
			throw new Error("element " + query + " "+e);
		}
		if (!elm)
			throw new Error("element " + query + " is not found.");
		this.element = elm;
		return elm;
	}
	async queryElements(query) {
		let elms = null;
		try {
			var q = this.parseQuery(query);
			elms = await this.driver.findElements(q);
		}
		catch(e) {
			throw new Error("elements " + query + " "+e);
		}
		if (!elms)
			throw new Error("elements " + query + " is not found.");
		return elms;
	}
	async waitElement(query, timeout) {
		if(query === null || query === undefined) {
			if (this.element !== null)
				return this.element;
		}
		if(query instanceof WebElement) {
			this.element = query;
			return query;
		}
		let elm = null;
		let wait = 1000;
		if(timeout)
			wait = timeout;
		try {
			elm = await this.driver.wait(until.elementLocated(this.parseQuery(query)), wait);
		}
		catch(e) {
			throw new Error("element " + query + " "+e);
		}
		if (!elm)
			throw new Error("element " + query + " is not found.");
		this.element = elm;
		return elm;
	}
	async waitVisible(query, timeout) {
		const elm = await this.waitElement(query,timeout);
		await this.driver.wait(until.elementIsVisible(elm),timeout);
		return elm;
	}
	async waitURL(pattern,timeout) {
		await this.driver.wait(until.urlMatches(pattern),timeout);
	}

	pauseAction(actions, msec) {
		return actions.pause(msec);
	}
	async pause(msec) {
		await this.pauseAction(this.driver.actions({bridge: true}),msec).perform();
	}
	async mouseMoveAction(actions, elm, query, position) {
		let x = 0;
		let y = 0;
		if (this.isPosition(position)) {
			const rect = await elm.getRect();
			x = position.x - rect.width / 2;
			y = position.y - rect.height / 2;
		}
		return await actions
			.move({origin: elm, x: x, y: y});
	}
	async mouseMove(query, position, timeout) {
		let elm = null;
		if(timeout)
			elm = await this.waitElement(query, timeout);
		else
			elm = await this.queryElement(query);
		const actions = await this.mouseMoveAction(this.driver.actions({bridge: true}), elm, query, position);
		await actions.perform();
		return elm;
	}

	async mouseDrag(query, position, button) {
		const elm = await this.queryElement(query);
		let fx = 0;
		let fy = 0;
		let tx = 0;
		let ty = 0;
		let rect = null;
		let b = Button.LEFT;
		if(this.isPosition(position)) {
			rect = await elm.getRect();
			fx = position.x - rect.width / 2;
			fy = position.y - rect.height / 2;
		}
		if(this.isToPosition(position)) {
			if(rect === null)
				rect = await elm.getRect();
			tx = position.tx - rect.width / 2;
			ty = position.ty - rect.height / 2;
		}
		if(button)
			b = button;
		let actions = this.driver.actions({bridge: true});
		await actions
			.move({origin: elm, x: fx, y: fy})
			.press(b)
			.move({origin: elm, x: tx, y: ty})
			.release(b)
			.perform();
		return elm;
	}

	async mouseSwipe(query, direction) {
		const elm = await this.queryElement(query);
		const rect = await elm.getRect();
		let x1 = 0;
		let y1 = 0;
		let x2 = 0;
		let y2 = 0;
		if(direction === "TtoB") {
			y1 = Math.floor(-rect.height * 0.3);
			y2 = Math.floor(rect.height * 0.3);
			if(y1 === y2 )
				y2 = y1+1;
		} else if(direction === "BtoT") {
			y1 = Math.floor(rect.height * 0.3);
			y2 = Math.floor(-rect.height * 0.3);
			if(y1 === y2 )
				y1 = y2+1;
		}
		else if(direction === "RtoL") {
			x1 = Math.floor(rect.width * 0.3);
			x2 = Math.floor(-rect.width * 0.3);
			if (x1 === x2)
				x1 = x2 + 1;
		}
		else { //if(direction === "LtoR") {
			x1 = Math.floor(-rect.width * 0.3);
			x2 = Math.floor(rect.width * 0.3);
			if (x1 === x2)
				x2 = x1 + 1;
		}
		const actions = this.driver.actions({bridge: true});
		await actions
			.move({origin: elm, x: x1, y: y1})
			.press()
			.move({origin: elm, x: x2, y: y2})
			.release()
			.perform();
		return elm;
	}

	async mouseClickAction(actions, elm, query, offset) {
		if (query) {
			if (this.isPosition(offset)) {
				const rect = await elm.getRect();
				const x = offset.x - rect.width / 2;
				const y = offset.y - rect.height / 2;
				return actions
					.move({origin: elm, x: x, y: y})
					.click();
			} else {
				return actions
					.move({origin: elm})
					.click();
			}
		} else {
			if (this.isPosition(offset)) {
				return actions
					.move({x: offset.x, y: offset.y})
					.click();
			} else {
				return actions
					.click();
			}
		}
		return actions;
	}

	async mouseClick(query, offset, timeout) {
		let elm = null;
		if(timeout)
			elm = await this.waitElement(query, timeout);
		else
			elm = await this.queryElement(query);
		const actions = await this.mouseClickAction(this.driver.actions({bridge: true}), elm, query, offset);
		await actions.perform();
		return elm;
	}

	async mousePressAction(actions, elm, query, offset, button) {
		if (query) {
			let x = 0;
			let y = 0;
			if (this.isPosition(offset)) {
				const rect = await elm.getRect();
				x = offset.x - rect.width / 2;
				y = offset.y - rect.height / 2;
			}
			actions = actions.move({
				origin: elm,
				x: x,
				y: y
			});
		}
		let b = Button.LEFT;
		if(button)
			b = button;
		return actions.press(b);
	}

	async mousePress(query, offset, button, timeout) {
		let elm = null;
		if(timeout)
			elm = await this.waitElement(query, timeout);
		else
			elm = await this.queryElement(query);
		const actions = await this.mousePressAction(this.driver.actions({bridge: true}), elm, query, offset, button);
		await actions.perform();
		return elm;
	}

	async mouseReleaseAction(actions, elm, query, offset, button) {
		if (query) {
			let x = 0;
			let y = 0;
			if (offset) {
				const rect = await elm.getRect();
				x = offset.x - rect.width / 2;
				y = offset.y - rect.height / 2;
			}
			actions = actions.move({
				origin: elm,
				x: x,
				y: y
			});
		}
		let b = Button.LEFT;
		if(button)
			b = button;
		return actions.release(b);
	}
	async mouseRelease(query, offset, button, timeout) {
		let elm = null;
		if(timeout)
			elm = await this.waitElement(query, timeout);
		else
			elm = await this.queryElement(query);
		const actions = await this.mouseReleaseAction(this.driver.actions({bridge: true}), elm, query, offset, button);
		await actions.perform();
		return elm;
	}
	async longPress(query, offset, button, wait) {
		let elm = null;
		let actions = this.driver.actions({bridge: true});
		if (query) {
			elm = await this.queryElement(query);
			const rect = await elm.getRect();
			let x = 0;
			let y = 0;
			if (this.isPosition(offset)) {
				x = offset.x - rect.width / 2;
				y = offset.y - rect.height / 2;
			}
			actions = actions.move({
				origin: elm,
				x: x,
				y: y
			});
		}
		let b = Button.LEFT;
		if(button)
			b = button;

		let wt = 1200;
		if(wait && wait > 0)
			wt = wait;
		await actions
			.press(b)
			.perform();
		await this.delay(wt);
		await actions
			.release(b)
			.perform();
		return elm;
	}

	async sendKeys(query, keys) {
		const elm = await this.queryElement(query);
		await elm.sendKeys(...keys);
		return elm;
	}

	keyDownAction(actions, key) {
		return actions.keyDown(key);
	}
	async keyDown(key) {
		await this.keyDownAction(this.driver.actions({bridge: true}),key).perform();
	}
	keyUpAction(actions, key) {
		return actions.keyUp(key);
	}
	async keyUp(key) {
		await this.keyUpAction(this.driver.actions({bridge: true}),key).perform();
	}

	async executeScript(script) {
		return await this.driver.executeScript(script);
	}

	async waitAlert(timeout) {
		if(timeout)
			await this.driver.wait(until.alertIsPresent(), timeout);
		else
			await this.driver.wait(until.alertIsPresent());
	}
	async getAlertText() {
		let alert = await this.driver.switchTo().alert();
		return await alert.getText();
	}
	async setAlertText(keys) {
		let alert = await this.driver.switchTo().alert();
		await alert.sendKeys(...keys);
	}
	async acceptAlert() {
		let alert = await this.driver.switchTo().alert();
		await alert.accept();
	}
	async dismissAlert() {
		let alert = await this.driver.switchTo().alert();
		await alert.dismiss();
	}

	async getElementText(query) {
		const elm = await this.queryElement(query);
		const text =  await elm.getText();
		return text;
	}
	async isElementSelected(query) {
		const elm = await this.queryElement(query);
		const state =  await elm.isSelected();
		return state;
	}
	async getElementAttribute(query, attr) {
		const elm = await this.queryElement(query);
		const state =  await elm.getAttribute(attr);
		return state;
	}
	saveImage(img_b64, filePath) {
		const img_bin = Buffer.from(img_b64, "base64");
		const fnDir = path.dirname(filePath);
		if (!fs.existsSync(fnDir))
			fs.mkdirSync(fnDir,{recursive:true});
		fs.writeFileSync(filePath, img_bin);
	}
	async takeScreenshot(cond) {
		let img_b64 = await this.driver.takeScreenshot();
		if(cond.query) {
			const elm = await this.queryElement(cond.query);
			const rect = await elm.getRect();
			const img_bin = Buffer.from(img_b64, "base64");
			const cropPosition = {left:rect.x*2, top:rect.y*2, width:rect.width*2, height:rect.height*2};
			await sharp(img_bin)
				.extract(cropPosition)
				.toBuffer()
				.then(data=>{img_b64 = data.toString("base64");});
		}
		if(this.isRect(cond.rect)) {
			const img_bin = Buffer.from(img_b64, "base64");
			await sharp(img_bin)
				.extract(cond.rect)
				.toBuffer()
				.then(data=>{img_b64 = data.toString("base64");});
		}
		if(cond.compare) {
			const hashHex = crypto.createHash('sha256').update(img_b64, 'utf8').digest('hex');
			if(hashHex !== cond.compare) {
				if(cond.saveOnError)
				try {
					this.saveImage(img_b64, cond.saveOnError);
				}
				catch(ex) {
					console.log(ex);
				}
				throw new Error("screenshot compare error hash=" + hashHex);
			}
		}
		if(cond.save) {
			this.saveImage(img_b64,cond.save);
		}
		return img_b64;
	}
}